package traffic;
import java.util.*;
import java.util.concurrent.*;

/**
 * 路口
 * 	负责每个路口控制车辆随机生成以及3个方向车辆（直行，左行，右行）的通行
 * @author lihao
 */

public class Road {
	/** 路口名称 */
	private String name;
	
	/** 执行车辆生成和车辆通行任务的线程池 */
	ExecutorService exec = Executors.newCachedThreadPool();

	/** 直行车队列 */
	private BlockingQueue<Vehicle> straightVehicles = new LinkedBlockingQueue<Vehicle>();
	/** 左行车队列 */
	private BlockingQueue<Vehicle> leftVehicles = new LinkedBlockingQueue<Vehicle>();
	/** 右行车队列 */
	private BlockingQueue<Vehicle> rightVehicles = new LinkedBlockingQueue<Vehicle>();

	public Road(String name, TrafficLamp leftLamp, TrafficLamp straightLamp) {
		this.name = name;
		// 开启车辆生成线程
		exec.execute(new VehicleGenerator());
		// 开启3个方向通行线程
		exec.execute(new VehiclesPass(leftLamp, leftVehicles));
		exec.execute(new VehiclesPass(TrafficLamp.RIGHT, rightVehicles));
		exec.execute(new VehiclesPass(straightLamp, straightVehicles));
	}
	
	/** 关闭路口线程 */
	public void shutdown() {
		exec.shutdownNow();
	}
	
	
	/** 车辆生成任务。封装为Runnable对象 */
	private class VehicleGenerator implements Runnable {
		private Random rand = new Random(); // 随机车辆生成间隔
		@Override
		public void run() {
			try {
				while (!Thread.interrupted()) {
					TimeUnit.MILLISECONDS.sleep(rand.nextInt(500));
					// 随机生成车辆行驶方向，并放入相应车辆队列
					int direction = rand.nextInt(3);
					switch (direction) {
					case 0:
						leftVehicles.put(new Vehicle(Direction.valueOf("LEFT")));
						break;
					case 1:
						rightVehicles.put(new Vehicle(Direction.valueOf("RIGHT"))); 
						break; 
					case 2: 
						straightVehicles.put(new Vehicle(Direction.valueOf("STRAIGHT")));
						break;
					}
				}
			} catch (InterruptedException e) {
				System.out.println("VehicleGenerator interrupted");
			}
		}
		
	}
	
	/** 车辆通行任务。封装为Runnable对象 */
	private class VehiclesPass implements Runnable {
		private TrafficLamp lamp;
		private BlockingQueue<Vehicle> vehicles;
		public VehiclesPass(TrafficLamp lamp, BlockingQueue<Vehicle> vehicles) {
			this.lamp = lamp;
			this.vehicles = vehicles;
		}
		@Override
		public void run() {
			try {
				while (!Thread.interrupted()) {
					TimeUnit.SECONDS.sleep(1); // 模拟每一秒通行一辆车
					// 若当前灯可通行，则让相应车辆队列的队首车辆通行
					if (lamp.isPassable()) {
						Vehicle v = null;
						if ((v = vehicles.poll()) != null) {
							System.out.println(Road.this.name + ": " + v);
						}
					}
				}
			} catch (InterruptedException e) {
				System.out.println(Road.this.name + ":" + lamp + " interrupted");
			}
		}
		
	}
}
